/*
 * Copyright 2010-2017, Tarantool AUTHORS, please see AUTHORS file.
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above
 *    copyright notice, this list of conditions and the
 *    following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY <COPYRIGHT HOLDER> ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * <COPYRIGHT HOLDER> OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Main file for the SQLite library.  The routines in this file
 * implement the programmer interface to the library.  Routines in
 * other files are for internal use by SQLite and should not be
 * accessed by users of the library.
 */

#include "sqliteInt.h"
#include "box/session.h"

/*
 * Execute SQL code.  Return one of the SQLITE_ success/failure
 * codes.  Also write an error message into memory obtained from
 * malloc() and make *pzErrMsg point to that message.
 *
 * If the SQL is a query, then for each row in the query result
 * the xCallback() function is called.  pArg becomes the first
 * argument to xCallback().  If xCallback=NULL then no callback
 * is invoked, even for queries.
 */
int
sqlite3_exec(sqlite3 * db,	/* The database on which the SQL executes */
	     const char *zSql,	/* The SQL to be executed */
	     sqlite3_callback xCallback,	/* Invoke this callback routine */
	     void *pArg,	/* First argument to xCallback() */
	     char **pzErrMsg	/* Write error messages here */
    )
{
	int rc = SQLITE_OK;	/* Return code */
	const char *zLeftover;	/* Tail of unprocessed SQL */
	sqlite3_stmt *pStmt = 0;	/* The current SQL statement */
	char **azCols = 0;	/* Names of result columns */
	int callbackIsInit;	/* True if callback data is initialized */
	struct session *user_session = current_session();

	if (!sqlite3SafetyCheckOk(db))
		return SQLITE_MISUSE_BKPT;
	if (zSql == 0)
		zSql = "";

	sqlite3Error(db, SQLITE_OK);
	while (rc == SQLITE_OK && zSql[0]) {
		int nCol;
		char **azVals = 0;

		pStmt = 0;
		rc = sqlite3_prepare_v2(db, zSql, -1, &pStmt, &zLeftover);
		assert(rc == SQLITE_OK || pStmt == 0);
		if (rc != SQLITE_OK) {
			continue;
		}
		if (!pStmt) {
			/* this happens for a comment or white-space */
			zSql = zLeftover;
			continue;
		}

		callbackIsInit = 0;
		nCol = sqlite3_column_count(pStmt);

		while (1) {
			int i;
			rc = sqlite3_step(pStmt);
			/* Invoke the callback function if required */
			if (xCallback && (SQLITE_ROW == rc ||
					  (SQLITE_DONE == rc && !callbackIsInit
					   && user_session->
					   sql_flags & SQLITE_NullCallback))) {
				if (!callbackIsInit) {
					azCols =
					    sqlite3DbMallocZero(db,
								2 * nCol *
								sizeof(const
								       char *) +
								1);
					if (azCols == 0) {
						goto exec_out;
					}
					for (i = 0; i < nCol; i++) {
						azCols[i] =
						    (char *)
						    sqlite3_column_name(pStmt,
									i);
						/* sqlite3VdbeSetColName() installs column names as UTF8
						 * strings so there is no way for sqlite3_column_name() to fail.
						 */
						assert(azCols[i] != 0);
					}
					callbackIsInit = 1;
				}
				if (rc == SQLITE_ROW) {
					azVals = &azCols[nCol];
					for (i = 0; i < nCol; i++) {
						azVals[i] =
						    (char *)
						    sqlite3_column_text(pStmt,
									i);
						if (!azVals[i]
						    &&
						    sqlite3_column_type(pStmt,
									i) !=
						    SQLITE_NULL) {
							sqlite3OomFault(db);
							goto exec_out;
						}
					}
				}
				if (xCallback(pArg, nCol, azVals, azCols)) {
					/* EVIDENCE-OF: R-38229-40159 If the callback function to
					 * sqlite3_exec() returns non-zero, then sqlite3_exec() will
					 * return SQLITE_ABORT.
					 */
					rc = SQLITE_ABORT;
					sqlite3VdbeFinalize((Vdbe *) pStmt);
					pStmt = 0;
					sqlite3Error(db, SQLITE_ABORT);
					goto exec_out;
				}
			}

			if (rc != SQLITE_ROW) {
				rc = sqlite3VdbeFinalize((Vdbe *) pStmt);
				pStmt = 0;
				zSql = zLeftover;
				while (sqlite3Isspace(zSql[0]))
					zSql++;
				break;
			}
		}

		sqlite3DbFree(db, azCols);
		azCols = 0;
	}

 exec_out:
	if (pStmt)
		sqlite3VdbeFinalize((Vdbe *) pStmt);
	sqlite3DbFree(db, azCols);

	rc = sqlite3ApiExit(db, rc);
	if (rc != SQLITE_OK && pzErrMsg) {
		int nErrMsg = 1 + sqlite3Strlen30(sqlite3_errmsg(db));
		*pzErrMsg = sqlite3Malloc(nErrMsg);
		if (*pzErrMsg) {
			memcpy(*pzErrMsg, sqlite3_errmsg(db), nErrMsg);
		} else {
			rc = SQLITE_NOMEM_BKPT;
			sqlite3Error(db, SQLITE_NOMEM);
		}
	} else if (pzErrMsg) {
		*pzErrMsg = 0;
	}

	assert((rc & db->errMask) == rc);
	return rc;
}
